//  PyCOM - Distributed Component Model for Python - C++ Client Library
//  Copyright (c) 2011-2012, Dmitry "Divius" Tantsur
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following disclaimer
//    in the documentation and/or other materials provided with the
//    distribution.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <memory>
#include <stdexcept>

#include <zmq.hpp>

#include <pycom.h>

#include "pycom-private.h"

namespace pycom {

static const char *NS_IFACE_NAME = "org.pycom.nameserver";
static const char *NS_LOCATE_METHOD = "locate";

Context::Context(const std::string &ns_address) : _pvt(new _context_pvt) {
    if (ns_address.empty() || ns_address.substr(0, 6) != "tcp://") {
        throw std::invalid_argument("Malformed or empty ns_address");
    }
    _pvt->ns_address = ns_address;
}

Context::~Context() {
    delete _pvt;
}

Component *Context::nameserver() {
    return open(NS_IFACE_NAME, _pvt->ns_address);
}

Component *Context::locate(const std::string &iface_name) {
    std::auto_ptr<Component> ns(nameserver());
    Value temp;

    temp["interface"] = iface_name;

    try {
        temp = ns->invoke(NS_LOCATE_METHOD, temp);
    }
    catch (RemoteError &err) {
        if (err.code() == ERROR_SERVICE_NOT_FOUND) {
            throw LookupError("Service cannot be found: " + iface_name);
        }
        else {
            throw;
        }
    }

    return open(iface_name, temp["address"].asString());
}

// Private

Component *Context::open(const std::string &iface_name, const std::string &address) {
    _component_pvt *component_pvt = new _component_pvt(_pvt, iface_name);
    try {
        component_pvt->zsock.connect(address.c_str());
    }
    catch (zmq::error_t) {
        int linger = 0;
        component_pvt->zsock.setsockopt(ZMQ_LINGER, &linger, sizeof(linger));
        delete component_pvt;
        throw LookupError("Cannot connect to server");
    }

    return new Component(component_pvt);
}

}
